﻿__copyright__ = "Copyright (c) 2017 CollabNet, Inc. All rights reserved."

from app.scripts import BaseInitializationScript

import app.util
import os
import re

class InitializationScript(BaseInitializationScript):

    scriptDescription = "Setup programmatic MongoDB users"

    def initialize(self):

        # NOTE: this code takes advantage of Mongo's `localhost exception` to
        # create the programmatic users at startup, with authentication enabled
        # by configuration.  This permits the first user - the admin user - to
        # be created via localhost interface without authentication, so long as
        # there are no other users in mongo.  Subsequent users have to be added
        # using the admin credentials.
        #
        # This code does NOT update credentials, if they are modified in config.

        admin_creds = {
            'username': self.cfg['MONGODB_APP_ADMIN_USER'],
            'password': self.cfg['MONGODB_APP_ADMIN_PASSWORD'],
            'db':       self.cfg['MONGODB_ADMIN_DATABASE_NAME'],
            'roles':    [ 'userAdminAnyDatabase', 'readWriteAnyDatabase',
                          'dbAdminAnyDatabase' ]
        }

        backup_creds = {
            'username': self.cfg['MONGODB_APP_BACKUP_USER'],
            'password': self.cfg['MONGODB_APP_BACKUP_PASSWORD'],
            'db':       self.cfg['MONGODB_APP_DATABASE_NAME'],
            'roles':    ['userAdmin', 'readWrite', 'dbAdmin']
        }

        appuser_creds = {
            'username': self.cfg['MONGODB_APP_SERVICES_USER'],
            'password': self.cfg['MONGODB_APP_SERVICES_PASSWORD'],
            'db':       self.cfg['MONGODB_APP_DATABASE_NAME'],
            'roles':    ['readWrite', 'dbAdmin']
        }

        self.existing_passwords = self.readExsitingPasswords()

        self.addOrUpdateMongoUser(admin_creds)
        self.addOrUpdateMongoUser(backup_creds, admin_creds)
        self.addOrUpdateMongoUser(appuser_creds, admin_creds)

    def addOrUpdateMongoUser(self, creds, admin_creds=None):
        db, user, password, roles = creds['db'], creds['username'], creds['password'], creds['roles']

        self.log.info("Checking for existing user %s in MongoDB" % user)

        if not self.mongodbUserExists(db, user, password):
            self.log.info("Adding MongoDB user %s" % user)
            role_str = "['" + "','".join(str(x) for x in roles) + "']"
            auth_str = ''
            if admin_creds:
                auth_str = "%s -u '%s' -p '%s'" % (admin_creds['db'], admin_creds['username'], admin_creds['password'])
            command = "/usr/bin/mongo %s --eval \"db.getSiblingDB('%s').createUser({user: '%s', pwd: '%s', roles: %s})\"" % (auth_str, db, user, password, role_str)
            app.util.executeCommand(command)
        else:
            self.log.info("MongoDB user %s already exists" % user)
            self.updatePassword(db, user, password, admin_creds)
        self.__writePasswordsFile(user, password)

    def mongodbUserExists(self, db, user, password):
        cmd_str = "/usr/bin/mongo %s -u '%s' -p '%s' --eval \"db.getName()\""
        
        if user in self.existing_passwords:
            try:
                output = app.util.executeCommand(cmd_str % (db, user, self.existing_passwords[user]))
                self.log.info("User %s located with prior credentials")
                return re.search('^%s$' % db, output, re.MULTILINE)
            except:
                pass
        try:
            output = app.util.executeCommand(cmd_str % (db, user, password))
            self.log.info("User %s located with current credentials")
            return re.search('^%s$' % db, output, re.MULTILINE)
        except:
            return False        

    def updatePassword(self, db, user, password, admin_creds=None):

        self.log.info("Updating password for %s in MongoDB" % user)

        if admin_creds:
            self.log.info("Using suplied administrative credentials for update")
            auth_str = "%s -u '%s' -p '%s'" % (admin_creds['db'], 
                                               admin_creds['username'], 
                                               admin_creds['password'])
        elif user in self.existing_passwords:
            self.log.info("Using discovered administrative credentials for update")
            auth_str = "%s -u '%s' -p '%s'" % (self.cfg['MONGODB_ADMIN_DATABASE_NAME'],
                                               self.cfg['MONGODB_APP_ADMIN_USER'],
                                               self.existing_passwords[user])
        else:             
            self.log.info("Using default administrative credentials for update")
            auth_str = "%s -u '%s' -p '%s'" % (self.cfg['MONGODB_ADMIN_DATABASE_NAME'],
                                               self.cfg['MONGODB_APP_ADMIN_USER'], 
                                               self.cfg['MONGODB_APP_ADMIN_PASSWORD'])

        if db == 'admin':
            change_schema_command = "/usr/bin/mongo %s --eval \"db.getSiblingDB('%s').runCommand({authSchemaUpgrade: 3 })\"" % (auth_str, db)
            app.util.executeCommand(change_schema_command)

        command = "/usr/bin/mongo %s --eval \"db.getSiblingDB('%s').changeUserPassword('%s', '%s')\"" % (auth_str, db, user, password)
        app.util.executeCommand(command)

    def readExsitingPasswords(self):
        password_file = os.path.expanduser(self.cfg['MONGODB_PASSWORDS_FILE'])
        passwords = {}
        if os.path.isfile(password_file):
            with open(password_file) as file_data:
                for line in file_data:
                    m = re.match('^\s*(\w+)\s*:\s*(\S*)\s*', line)
                    if m:
                        passwords[m.group(1)] = m.group(2)
        return passwords            

    def __writePasswordsFile(self, user, password):
        password_file = self.cfg['MONGODB_PASSWORDS_FILE']
        tmp_password_file = password_file + '.tmp'        
        app.util.runSystem("touch %s" % password_file, self.log)
        app.util.runSystem("sed '/^%s:.*/d' %s > %s" % (
            user, 
            password_file, 
            tmp_password_file), self.log)
        command = "echo -e \"%s: %s\" >> %s" % (
            user, 
            password,
            tmp_password_file
        )
        app.util.runSystem(command, self.log)
        app.util.runSystem("mv %s %s" % (tmp_password_file, password_file), self.log)
        app.util.runSystem("chmod 600 %s" % password_file, self.log)
